//============================================================
//
//  section : FFT Cooley-Tukey / FoxLima filter
//
//            ________________________________________________
//
//                  71.log(N)    ns/sample FFT (double)
//               0.083.N^0.79    ns/sample FoxLima -"-
//
//              Proc :x86_64 Intel Xeon E312xx Sandy Bridge
//                    2.6 GHz
//              Cache: 32k, 32k, 4096k
//              Linux: 2.6.32-504.16.2.el6.x86_64
//              Gcc  : 4.4.7 20120313 (Red Hat 4.4.7-11)
//            ________________________________________________
//
//            constructor: fxl F(T*, T*, N,  q, str, ...) 
//                               ^   ^   ^   ^    ^    ^ 
//                               |   |   |   |    |    |_ arg
//                               |   |   |   |    |
//                               |   |   |   |    |_ command
//                               |   |   |   |
//                               |   |   |   |_ 2^q FFT bins
//                               |   |   |      (allows to 
//                               |   |   |       oversample) 
//                               |   |   |
//                               |   |   |_ # i/p samples
//                               |   |
//                               |   |_ pointer "real"-part
//                               |      (symm. for mtx)
//                               |
//                               |_ pointer "imag"-part
//                                  (antisymm. for mtx)
//
//                        oversampling: p = 2^q / N
//
//                        recommended q > 3 + log(N)/log(2)
//
//                        - command = "F0" , "F0  raw"
//                                    "F1"
//                                    "F2" , "F2  raw"
//                                    "F4" , "F4  raw"
//                                    "F8" , "F8  raw"
//                                    "F16", "F16 raw"
//                                    "GX" , "GX  raw"
//                                    "FC" , "FC  raw"
//                                    "FX" , "FX  raw"
//                                    "FE" , "FE  raw"
//                                    "FW" , "FW  raw"
//
//                          F0  = Cooley-Tukey, -6dB/oct
//                                            p = 2^q / N
//
//                          F1  = FoxLima-F1,   -6dB/oct
//                                            p = 2^q / N
//
//                          F2  = FoxLima-F2,  -12dB/oct
//                                            p = 2^q / N
//                                similar to Welch apodisation,
//                                but with exact cancellation of
//                                1/(f-f0) tails. 
//
//                          F4  = FoxLima-F4,  -24dB/oct
//                                            p = 2^q / N
//                                a suis-generis Welch-4 window
//                                but with exact cancellation of
//                                1/(f-f0)^k tails (k = 1, 2, 3). 
//
//                          F8  = FoxLima-F8,  -48dB/oct
//                                            p = 2^q / N
//                                a suis-generis Welch-8 window
//                                -48dB/oct
//
//                          F16 = FoxLima-F16, -96dB/oct
//                                            p = 2^q / N
//                                a suis-generis Welch-16 window
//                                -96dB/oct
//
//                          GX  = gaussian
//                                            p = 2^q / N
//                                w/ gaussian(sigma) window
//
//                          FC  = coherent chirp
//                                            p = 2^q 
//                                Cooley-Tukey, no apodisation
//
//                          FX  = coherent chirp
//                                            p = 2^q 
//                                w/ gaussian(sigma) window
//
//                          FE  = coherent chirp
//                                            p = 2^q 
//                                w/ back-exp. (tau) window
//
//                          FW  = coherent chirp
//                                            p = 2^q 
//                                w/ Welch(sigma) window
//
//                  case   F0, F1, F2, F4, F8, F16 raw
//                     | - arg = void
//
//                  case  F0, F1, F2, F4, F8, F16 smooth
//                     | - arg = L3, L20
//                     |
//                     |         L3 *p = # bins at which 
//                     |                   smooth function
//                     |                   falls 3 dB
//                     |
//                     |         L20*p = # bins at which 
//                     |                   smooth function
//                     |                   falls 20 dB
//                     |
//                     |   1.472 < L20/L3 < 4.684 
//
//                  case   GX raw
//                     | - arg = sigma
//                     |
//                     |         sigma = # bins sigma
//
//                  case   GX smooth
//                     | - arg = sigma, L3, L20
//                     |
//                     |         L3 *p = # bins at which 
//                     |                   smooth function
//                     |                   falls 3 dB
//                     |
//                     |         L20*p = # bins at which 
//                     |                   smooth function
//                     |                   falls 20 dB
//                     |
//                     |         sigma = # bins sigma, for
//                     |                   option GX
//                     |
//                     |   1.472 < L20/L3 < 4.684 
//
//                  case   FC raw
//                     | - arg = 2*f0*delta, sigma
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |         sigma = # bins Cooley-window
//
//                  case   FX raw
//                     | - arg = 2*f0*delta, sigma
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |         sigma = # bins sigma
//
//                  case   FE raw
//                     | - arg = 2*f0*delta, tau
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |           tau = # bins tau
//
//                  case   FW raw
//                     | - arg = 2*f0*delta, sigma
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |         sigma = # bins sigma
//
//                  case   FC smooth
//                     | - arg = L3, L20, 2*f0*delta, sigma
//                     |
//                     |         L3 *p = # bins at which 
//                     |                   smooth function
//                     |                   falls 3 dB
//                     |
//                     |         L20*p = # bins at which 
//                     |                   smooth function
//                     |                   falls 20 dB
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |         sigma = # bins Cooley window
//                     |
//                     |   1.472 < L20/L3 < 4.684 
//
//                  case   FX smooth
//                     | - arg = L3, L20, 2*f0*delta, sigma
//                     |
//                     |         L3 *p = # bins at which 
//                     |                   smooth function
//                     |                   falls 3 dB
//                     |
//                     |         L20*p = # bins at which 
//                     |                   smooth function
//                     |                   falls 20 dB
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |         sigma = # bins sigma
//                     |
//                     |   1.472 < L20/L3 < 4.684 
//
//                  case   FE smooth
//                     | - arg = L3, L20, 2*f0*delta, tau
//                     |
//                     |         L3 *p = # bins at which 
//                     |                   smooth function
//                     |                   falls 3 dB
//                     |
//                     |         L20*p = # bins at which 
//                     |                   smooth function
//                     |                   falls 20 dB
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |           tau = # bins tau
//                     |
//                     |   1.472 < L20/L3 < 4.684 
//
//                  case   FW smooth
//                     | - arg = L3, L20, 2*f0*delta, sigma
//                     |
//                     |         L3 *p = # bins at which 
//                     |                   smooth function
//                     |                   falls 3 dB
//                     |
//                     |         L20*p = # bins at which 
//                     |                   smooth function
//                     |                   falls 20 dB
//                     |
//                     |    2*f0*delta = chirp coherence
//                     |                 freq. x sampling
//                     |                 delta x 2
//                     |               = f0 / FSR
//                     |
//                     |         sigma = # bins sigma
//                     |
//                     |   1.472 < L20/L3 < 4.684 
//
//
//            OUTPUT _________________________________________
//
//              fn   = (n-Np/2)/Np/T       df/f = 0.7 /int(fT) 
//
//              An   = sqrt(ddr^2+ddc^2)/Np  dA = that of i/p
//                   = sqrt(fzr^2+fzc^2)/Np
//
//                     _______________________________________
//
//              CASE-1 ...    [i/p units] = etc / sqrt(s)
// 
//              [o/p units] = [i/p units] * [i/p bin size]
//                                        * sqrt(Np)
//
//                     _______________________________________
//
//              CASE-2 ...    [i/p units] = etc
//
//              [o/p units] = [i/p units] * sqrt[i/p bin size]
//                                        * sqrt(Np)
//
//
//            CLASS PAR.'s ___________________________________
//
//               N   = actually Np
//               p   = oversampling ratio
//               rms = signal rms 
//
//               ddr = pointer to Np long array of "real" part
//                     (for matrices: "symmetrical" part)
//               ddc = pointer to Np long array of "imag" part
//                     (for matrices: -i * "anti-symmetrical")
//
//               fzr = pointer to Np long array of "real" part
//                     (for matrices: "symmetrical" part)
//                     after FoxLima filter
//               fzc = pointer to Np long array of "imag" part
//                     (for matrices: -i * "anti-symmetrical")
//                     after FoxLima filter
//
//               ddr/ddc = native          [ 0 ... +F -F ... 0]
//               fzr/fzc = segment-shifted [-F ...   0  ... +F]
//
//            ________________________________________________
//
//  code    : FoxLima filter FFT
//
//  author  : M. Dima
//            NXV4 Solutions
//
//  date    : FXL4            / Mon Jan 22 15:17:55 CEST 2018
//
//============================================================

#pragma once
#include <string>
#include <iostream>
#include <utility>
#include <type_traits>
#include <stdexcept>       // runtime_error

using std::cerr                                             ;
using std::endl                                             ;
using std::decay                                            ;
using std::is_arithmetic                                    ;
using std::runtime_error                                    ;



namespace fxl4
{

//===========================================================
//
 template<typename T>
 class fxl
   {public: typename decay<T>::type* ddr       = 0          ;
            typename decay<T>::type* ddc       = 0          ;
            typename decay<T>::type* fzr       = 0          ;
            typename decay<T>::type* fzc       = 0          ;
            int                      N         = 0          ;
            double                   p         = 0          ;
            double                   rms       = 0          ;


      template<typename... ARG>
      fxl  (T*, T*, int, int ,
            const std::string, ARG&& ... arg)               ;

      // ----------------------------------------------------

    private:

      bool raw = false                                      ;

      void foxlima(double, double, const std::string)       ;
      void cooley()                                         ;
      void f1()                                             ;
      void f2()                                             ;
      void f4()                                             ;
      void fc(double, double)                               ;
      void fx(double, double)                               ;
      void fe(double, double)                               ;
      void fw(double, double)                               ;

      // ----------------------------------------------------

    public:
     ~fxl()
      {if (ddr != 0) delete[] ddr                           ;
       if (ddc != 0) delete[] ddc                           ;
       if (fzr != 0) delete[] fzr                           ;
       if (fzc != 0) delete[] fzc                           ;}

 };
// 
//
//
//===========================================================
};
